@@ -3,12 +3,12 @@
  * print.c
  *	  various print routines (used mostly for debugging)
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/nodes/print.c,v 1.64 2003/08/11 23:04:49 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/nodes/print.c,v 1.91 2009/01/01 17:23:43 momjian Exp $
  *
  * HISTORY
  *	  AUTHOR			DATE			MAJOR EVENT
@@ -20,15 +20,13 @@
 #include "postgres.h"
 
 #include "access/printtup.h"
-#include "catalog/pg_type.h"
 #include "lib/stringinfo.h"
 #include "nodes/print.h"
 #include "optimizer/clauses.h"
 #include "parser/parsetree.h"
 #include "utils/lsyscache.h"
-#include "utils/syscache.h"
+#include "utils/rel.h"
 
-static char *plannode_type(Plan *p);
 
 /*
  * print
@@ -194,12 +192,20 @@ pretty_format_node_dump(const char *dump)
 					}
 					j = indentDist - 1;
 					/* j will equal indentDist on next loop iteration */
+					/* suppress whitespace just after } */
+					while (dump[i + 1] == ' ')
+						i++;
 					break;
 				case ')':
-					/* force line break after ')' */
-					line[j + 1] = '\0';
-					appendStringInfo(&str, "%s\n", line);
-					j = indentDist - 1;
+					/* force line break after ), unless another ) follows */
+					if (dump[i + 1] != ')')
+					{
+						line[j + 1] = '\0';
+						appendStringInfo(&str, "%s\n", line);
+						j = indentDist - 1;
+						while (dump[i + 1] == ' ')
+							i++;
+					}
 					break;
 				case '{':
 					/* force line break before { */
@@ -247,7 +253,7 @@ pretty_format_node_dump(const char *dump)
 void
 print_rt(List *rtable)
 {
-	List	   *l;
+	ListCell   *l;
 	int			i = 1;
 
 	printf("resno\trefname  \trelid\tinFromCl\n");
@@ -266,10 +272,6 @@ print_rt(List *rtable)
 				printf("%d\t%s\t[subquery]",
 					   i, rte->eref->aliasname);
 				break;
-			case RTE_FUNCTION:
-				printf("%d\t%s\t[rangefunction]",
-					   i, rte->eref->aliasname);
-				break;
 			case RTE_JOIN:
 				printf("%d\t%s\t[join]",
 					   i, rte->eref->aliasname);
@@ -278,6 +280,18 @@ print_rt(List *rtable)
 				printf("%d\t%s\t[special]",
 					   i, rte->eref->aliasname);
 				break;
+			case RTE_FUNCTION:
+				printf("%d\t%s\t[rangefunction]",
+					   i, rte->eref->aliasname);
+				break;
+			case RTE_VALUES:
+				printf("%d\t%s\t[values list]",
+					   i, rte->eref->aliasname);
+				break;
+			case RTE_CTE:
+				printf("%d\t%s\t[cte]",
+					   i, rte->eref->aliasname);
+				break;
 			default:
 				printf("%d\t%s\t[unknown rtekind]",
 					   i, rte->eref->aliasname);
@@ -325,7 +339,7 @@ print_expr(Node *expr, List *rtable)
 					RangeTblEntry *rte;
 
 					Assert(var->varno > 0 &&
-						   (int) var->varno <= length(rtable));
+						   (int) var->varno <= list_length(rtable));
 					rte = rt_fetch(var->varno, rtable);
 					relname = rte->eref->aliasname;
 					attname = get_rte_attribute_name(rte, var->varattno);
@@ -337,9 +351,8 @@ print_expr(Node *expr, List *rtable)
 	else if (IsA(expr, Const))
 	{
 		Const	   *c = (Const *) expr;
-		HeapTuple	typeTup;
 		Oid			typoutput;
-		Oid			typelem;
+		bool		typIsVarlena;
 		char	   *outputstr;
 
 		if (c->constisnull)
@@ -348,19 +361,10 @@ print_expr(Node *expr, List *rtable)
 			return;
 		}
 
-		typeTup = SearchSysCache(TYPEOID,
-								 ObjectIdGetDatum(c->consttype),
-								 0, 0, 0);
-		if (!HeapTupleIsValid(typeTup))
-			elog(ERROR, "cache lookup failed for type %u", c->consttype);
-		typoutput = ((Form_pg_type) GETSTRUCT(typeTup))->typoutput;
-		typelem = ((Form_pg_type) GETSTRUCT(typeTup))->typelem;
-		ReleaseSysCache(typeTup);
+		getTypeOutputInfo(c->consttype,
+						  &typoutput, &typIsVarlena);
 
-		outputstr = DatumGetCString(OidFunctionCall3(typoutput,
-													 c->constvalue,
-											   ObjectIdGetDatum(typelem),
-													 Int32GetDatum(-1)));
+		outputstr = OidOutputFunctionCall(typoutput, c->constvalue);
 		printf("%s", outputstr);
 		pfree(outputstr);
 	}
@@ -370,7 +374,7 @@ print_expr(Node *expr, List *rtable)
 		char	   *opname;
 
 		opname = get_opname(e->opno);
-		if (length(e->args) > 1)
+		if (list_length(e->args) > 1)
 		{
 			print_expr(get_leftop((Expr *) e), rtable);
 			printf(" %s ", ((opname != NULL) ? opname : "(invalid operator)"));
@@ -387,7 +391,7 @@ print_expr(Node *expr, List *rtable)
 	{
 		FuncExpr   *e = (FuncExpr *) expr;
 		char	   *funcname;
-		List	   *l;
+		ListCell   *l;
 
 		funcname = get_func_name(e->funcid);
 		printf("%s(", ((funcname != NULL) ? funcname : "(invalid function)"));
@@ -405,27 +409,36 @@ print_expr(Node *expr, List *rtable)
 
 /*
  * print_pathkeys -
- *	  pathkeys list of list of PathKeyItems
+ *	  pathkeys list of PathKeys
  */
 void
 print_pathkeys(List *pathkeys, List *rtable)
 {
-	List	   *i,
-			   *k;
+	ListCell   *i;
 
 	printf("(");
 	foreach(i, pathkeys)
 	{
-		List	   *pathkey = lfirst(i);
+		PathKey    *pathkey = (PathKey *) lfirst(i);
+		EquivalenceClass *eclass;
+		ListCell   *k;
+		bool		first = true;
+
+		eclass = pathkey->pk_eclass;
+		/* chase up, in case pathkey is non-canonical */
+		while (eclass->ec_merged)
+			eclass = eclass->ec_merged;
 
 		printf("(");
-		foreach(k, pathkey)
+		foreach(k, eclass->ec_members)
 		{
-			PathKeyItem *item = lfirst(k);
+			EquivalenceMember *mem = (EquivalenceMember *) lfirst(k);
 
-			print_expr(item->key, rtable);
-			if (lnext(k))
+			if (first)
+				first = false;
+			else
 				printf(", ");
+			print_expr((Node *) mem->em_expr, rtable);
 		}
 		printf(")");
 		if (lnext(i))
@@ -441,17 +454,17 @@ print_pathkeys(List *pathkeys, List *rtable)
 void
 print_tl(List *tlist, List *rtable)
 {
-	List	   *tl;
+	ListCell   *tl;
 
 	printf("(\n");
 	foreach(tl, tlist)
 	{
-		TargetEntry *tle = lfirst(tl);
+		TargetEntry *tle = (TargetEntry *) lfirst(tl);
 
-		printf("\t%d %s\t", tle->resdom->resno,
-			   tle->resdom->resname ? tle->resdom->resname : "<null>");
-		if (tle->resdom->ressortgroupref != 0)
-			printf("(%u):\t", tle->resdom->ressortgroupref);
+		printf("\t%d %s\t", tle->resno,
+			   tle->resname ? tle->resname : "<null>");
+		if (tle->ressortgroupref != 0)
+			printf("(%u):\t", tle->ressortgroupref);
 		else
 			printf("    :\t");
 		print_expr((Node *) tle->expr, rtable);
@@ -467,145 +480,16 @@ print_tl(List *tlist, List *rtable)
 void
 print_slot(TupleTableSlot *slot)
 {
-	if (!slot->val)
+	if (TupIsNull(slot))
 	{
 		printf("tuple is null.\n");
 		return;
 	}
-	if (!slot->ttc_tupleDescriptor)
+	if (!slot->tts_tupleDescriptor)
 	{
 		printf("no tuple descriptor.\n");
 		return;
 	}
 
-	debugtup(slot->val, slot->ttc_tupleDescriptor, NULL);
-}
-
-static char *
-plannode_type(Plan *p)
-{
-	switch (nodeTag(p))
-	{
-		case T_Plan:
-			return "PLAN";
-		case T_Result:
-			return "RESULT";
-		case T_Append:
-			return "APPEND";
-		case T_Scan:
-			return "SCAN";
-		case T_SeqScan:
-			return "SEQSCAN";
-		case T_IndexScan:
-			return "INDEXSCAN";
-		case T_TidScan:
-			return "TIDSCAN";
-		case T_SubqueryScan:
-			return "SUBQUERYSCAN";
-		case T_FunctionScan:
-			return "FUNCTIONSCAN";
-		case T_Join:
-			return "JOIN";
-		case T_NestLoop:
-			return "NESTLOOP";
-		case T_MergeJoin:
-			return "MERGEJOIN";
-		case T_HashJoin:
-			return "HASHJOIN";
-		case T_Material:
-			return "MATERIAL";
-		case T_Sort:
-			return "SORT";
-		case T_Agg:
-			return "AGG";
-		case T_Unique:
-			return "UNIQUE";
-		case T_SetOp:
-			return "SETOP";
-		case T_Limit:
-			return "LIMIT";
-		case T_Hash:
-			return "HASH";
-		case T_Group:
-			return "GROUP";
-		default:
-			return "UNKNOWN";
-	}
-}
-
-/*
- * Recursively prints a simple text description of the plan tree
- */
-void
-print_plan_recursive(Plan *p, Query *parsetree, int indentLevel, char *label)
-{
-	int			i;
-	char		extraInfo[NAMEDATALEN + 100];
-
-	if (!p)
-		return;
-	for (i = 0; i < indentLevel; i++)
-		printf(" ");
-	printf("%s%s :c=%.2f..%.2f :r=%.0f :w=%d ", label, plannode_type(p),
-		   p->startup_cost, p->total_cost,
-		   p->plan_rows, p->plan_width);
-	if (IsA(p, Scan) ||
-		IsA(p, SeqScan))
-	{
-		RangeTblEntry *rte;
-
-		rte = rt_fetch(((Scan *) p)->scanrelid, parsetree->rtable);
-		StrNCpy(extraInfo, rte->eref->aliasname, NAMEDATALEN);
-	}
-	else if (IsA(p, IndexScan))
-	{
-		RangeTblEntry *rte;
-
-		rte = rt_fetch(((IndexScan *) p)->scan.scanrelid, parsetree->rtable);
-		StrNCpy(extraInfo, rte->eref->aliasname, NAMEDATALEN);
-	}
-	else if (IsA(p, FunctionScan))
-	{
-		RangeTblEntry *rte;
-
-		rte = rt_fetch(((FunctionScan *) p)->scan.scanrelid, parsetree->rtable);
-		StrNCpy(extraInfo, rte->eref->aliasname, NAMEDATALEN);
-	}
-	else
-		extraInfo[0] = '\0';
-	if (extraInfo[0] != '\0')
-		printf(" ( %s )\n", extraInfo);
-	else
-		printf("\n");
-	print_plan_recursive(p->lefttree, parsetree, indentLevel + 3, "l: ");
-	print_plan_recursive(p->righttree, parsetree, indentLevel + 3, "r: ");
-
-	if (IsA(p, Append))
-	{
-		List	   *lst;
-		int			whichplan = 0;
-		Append	   *appendplan = (Append *) p;
-
-		foreach(lst, appendplan->appendplans)
-		{
-			Plan	   *subnode = (Plan *) lfirst(lst);
-
-			/*
-			 * I don't think we need to fiddle with the range table here,
-			 * bjm
-			 */
-			print_plan_recursive(subnode, parsetree, indentLevel + 3, "a: ");
-
-			whichplan++;
-		}
-	}
-}
-
-/* print_plan
-  prints just the plan node types */
-
-void
-print_plan(Plan *p, Query *parsetree)
-{
-	print_plan_recursive(p, parsetree, 0, "");
+	debugtup(slot, NULL);
 }
