@@ -266,12 +266,8 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 		if (AH->version < K_VERS_1_3)
 			die_horribly(AH, modulename, "direct database connections are not supported in pre-1.3 archives\n");
 
-		/*
-		 * We don't want to guess at whether the dump will successfully
-		 * restore; allow the attempt regardless of the version of the restore
-		 * target.
-		 */
-		AHX->minRemoteVersion = 0;
+		/* XXX Should get this from the archive */
+		AHX->minRemoteVersion = 070100;
 		AHX->maxRemoteVersion = 999999;
 
 		ConnectDatabase(AHX, ropt->dbname,
@@ -547,13 +543,7 @@ restore_toc_entry(ArchiveHandle *AH, TocEntry *te,
 
 					_selectOutputSchema(AH, "pg_catalog");
 
-					/* Send BLOB COMMENTS data to ExecuteSimpleCommands() */
-					if (strcmp(te->desc, "BLOB COMMENTS") == 0)
-						AH->outputKind = OUTPUT_OTHERDATA;
-
 					(*AH->PrintTocDataPtr) (AH, te, ropt);
-
-					AH->outputKind = OUTPUT_SQLCMDS;
 				}
 				else
 				{
@@ -569,8 +559,8 @@ restore_toc_entry(ArchiveHandle *AH, TocEntry *te,
 					/*
 					 * In parallel restore, if we created the table earlier in
 					 * the run then we wrap the COPY in a transaction and
-					 * precede it with a TRUNCATE.  If archiving is not on
-					 * this prevents WAL-logging the COPY.  This obtains a
+					 * precede it with a TRUNCATE.	If archiving is not on
+					 * this prevents WAL-logging the COPY.	This obtains a
 					 * speedup similar to that from using single_txn mode in
 					 * non-parallel restores.
 					 */
@@ -941,7 +931,7 @@ StartRestoreBlob(ArchiveHandle *AH, Oid oid)
 	/* Initialize the LO Buffer */
 	AH->lo_buf_used = 0;
 
-	ahlog(AH, 1, "restoring large object with OID %u\n", oid);
+	ahlog(AH, 2, "restoring large object with OID %u\n", oid);
 
 	if (AH->connection)
 	{
@@ -2419,7 +2409,7 @@ _doSetSessionAuth(ArchiveHandle *AH, const char *user)
 	appendPQExpBuffer(cmd, "SET SESSION AUTHORIZATION ");
 
 	/*
-	 * SQL requires a string literal here.  Might as well be correct.
+	 * SQL requires a string literal here.	Might as well be correct.
 	 */
 	if (user && *user)
 		appendStringLiteralAHX(cmd, user, AH);
@@ -2550,7 +2540,7 @@ _becomeUser(ArchiveHandle *AH, const char *user)
 }
 
 /*
- * Become the owner of the the given TOC entry object.  If
+ * Become the owner of the the given TOC entry object.	If
  * changes in ownership are not allowed, this doesn't do anything.
  */
 static void
@@ -3239,7 +3229,7 @@ restore_toc_entries_parallel(ArchiveHandle *AH)
 	}
 
 	/*
-	 * Now close parent connection in prep for parallel steps.  We do this
+	 * Now close parent connection in prep for parallel steps.	We do this
 	 * mainly to ensure that we don't exceed the specified number of parallel
 	 * connections.
 	 */
@@ -3739,7 +3729,7 @@ fix_dependencies(ArchiveHandle *AH)
 	/*
 	 * For some of the steps here, it is convenient to have an array that
 	 * indexes the TOC entries by dump ID, rather than searching the TOC list
-	 * repeatedly.  Entries for dump IDs not present in the TOC will be NULL.
+	 * repeatedly.	Entries for dump IDs not present in the TOC will be NULL.
 	 *
 	 * NOTE: because maxDumpId is just the highest dump ID defined in the
 	 * archive, there might be dependencies for IDs > maxDumpId.  All uses
@@ -3765,7 +3755,7 @@ fix_dependencies(ArchiveHandle *AH)
 	 * Note: currently, a TABLE DATA should always have exactly one
 	 * dependency, on its TABLE item.  So we don't bother to search, but look
 	 * just at the first dependency.  We do trouble to make sure that it's a
-	 * TABLE, if possible.  However, if the dependency isn't in the archive
+	 * TABLE, if possible.	However, if the dependency isn't in the archive
 	 * then just assume it was a TABLE; this is to cover cases where the table
 	 * was suppressed but we have the data and some dependent post-data items.
 	 */
@@ -3815,7 +3805,7 @@ fix_dependencies(ArchiveHandle *AH)
 
 	/*
 	 * It is possible that the dependencies list items that are not in the
-	 * archive at all.  Subtract such items from the depCounts.
+	 * archive at all.	Subtract such items from the depCounts.
 	 */
 	for (te = AH->toc->next; te != AH->toc; te = te->next)
 	{
