@@ -19,10 +19,10 @@
  * for example).
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- * $Id: pg_description.h,v 1.19 2003/08/04 02:40:12 momjian Exp $
+ * $PostgreSQL: pgsql/src/include/catalog/pg_description.h,v 1.28 2009/01/01 17:23:57 momjian Exp $
  *
  * NOTES
  *		the genbki.sh script reads this file and generates .bki
@@ -36,19 +36,16 @@
 #ifndef PG_DESCRIPTION_H
 #define PG_DESCRIPTION_H
 
-/* ----------------
- *		postgres.h contains the system type definitions and the
- *		CATALOG(), BOOTSTRAP and DATA() sugar words so this file
- *		can be read by both genbki.sh and the C compiler.
- * ----------------
- */
+#include "catalog/genbki.h"
 
 /* ----------------
  *		pg_description definition.	cpp turns this into
  *		typedef struct FormData_pg_description
  * ----------------
  */
-CATALOG(pg_description) BKI_WITHOUT_OIDS
+#define DescriptionRelationId  2609
+
+CATALOG(pg_description,2609) BKI_WITHOUT_OIDS
 {
 	Oid			objoid;			/* OID of object itself */
 	Oid			classoid;		/* OID of table containing object */
