@@ -1,4 +1,7 @@
 /*
+ * $PostgreSQL: pgsql/src/interfaces/libpq/win32.c,v 1.26 2009/06/11 14:49:14 momjian Exp $
+ *
+ *
  *	FILE
  *		win32.c
  *
@@ -12,26 +15,33 @@
  * The error constants are taken from the Frambak Bakfram LGSOCKET
  * library guys who in turn took them from the Winsock FAQ.
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  */
 
 /* Make stuff compile faster by excluding not used stuff */
 
-#define WIN32_LEAN_AND_MEAN
-#define WIN32_EXTRA_LEAN
 #define VC_EXTRALEAN
 #ifndef __MINGW32__
 #define NOGDI
 #endif
 #define NOCRYPT
 
-#include <windows.h>
-#include <winsock.h>
-#include <stdio.h>
+#include "postgres_fe.h"
+
 #include "win32.h"
 
+/* Declared here to avoid pulling in all includes, which causes name collissions */
+#ifdef ENABLE_NLS
+extern char *
+libpq_gettext(const char *msgid)
+__attribute__((format_arg(1)));
+#else
+#define libpq_gettext(x) (x)
+#endif
+
+
 static struct WSErrorEntry
 {
 	DWORD		error;
@@ -103,7 +113,7 @@ static struct WSErrorEntry
 		WSAEADDRINUSE, "Address already in use"
 	},
 	{
-		WSAEADDRNOTAVAIL, "Can't assign requested address"
+		WSAEADDRNOTAVAIL, "Cannot assign requested address"
 	},
 	{
 		WSAENETDOWN, "Network is down"
@@ -130,10 +140,10 @@ static struct WSErrorEntry
 		WSAENOTCONN, "Socket is not connected"
 	},
 	{
-		WSAESHUTDOWN, "Can't send after socket shutdown"
+		WSAESHUTDOWN, "Cannot send after socket shutdown"
 	},
 	{
-		WSAETOOMANYREFS, "Too many references, can't splice"
+		WSAETOOMANYREFS, "Too many references, cannot splice"
 	},
 	{
 		WSAETIMEDOUT, "Connection timed out"
@@ -289,7 +299,7 @@ winsock_strerror(int err, char *strerrbuf, size_t buflen)
 			dlls[i].handle = (void *) LoadLibraryEx(
 													dlls[i].dll_name,
 													0,
-											   LOAD_LIBRARY_AS_DATAFILE);
+													LOAD_LIBRARY_AS_DATAFILE);
 		}
 
 		if (dlls[i].dll_name && !dlls[i].handle)
@@ -302,14 +312,14 @@ winsock_strerror(int err, char *strerrbuf, size_t buflen)
 		success = 0 != FormatMessage(
 									 flags,
 									 dlls[i].handle, err,
-							   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
+								   MAKELANGID(LANG_ENGLISH, SUBLANG_DEFAULT),
 									 strerrbuf, buflen - 64,
 									 0
 			);
 	}
 
 	if (!success)
-		sprintf(strerrbuf, "Unknown socket error (0x%08X/%i)", err, err);
+		sprintf(strerrbuf, libpq_gettext("Unknown socket error (0x%08X/%i)"), err, err);
 	else
 	{
 		strerrbuf[buflen - 1] = '\0';
